/*
Contributors: Liam T. Monson
License: BSD 2-Clause license which should be found in the "license.txt"
file in the root game directory.
*/
#include "mainChess.hpp"
#include "gameStateProcessing.hpp"
#include "networking.hpp"

void cleanUp(struct chessGData &GData){
	glDeleteTextures(NUM_TEXTURES, texObjects);
	deInitNetworking(GData);
}

void checkForGameCompletion(struct chessGData &GData){
	if(isCheckMate(GData.board, GData.sMoves, GData.controlContext, GData.whiteTurn)){
		GData.controlContext = GAME_COMPLETION;
		GData.inGameMenuOpen = true;
	}
	else if(isDraw(GData.board, GData.sMoves, GData.controlContext, GData.whiteTurn)){
		GData.controlContext = GAME_COMPLETION;
	}	
}

void updateGame(struct chessGData &GData, GLFWwindow *window){
	checkForGameCompletion(GData);
	updateCursor(GData);
	updateCursorSelection(GData.menuSel, GData.selectedTile, GData.tCursX, GData.tCursY, GData.controlContext, GData.menuState);
	updateMoves(GData.board, GData.sMoves, GData.heldPieceLoc, GData.controlContext);
	updateAnimations(GData);
	//handle input
	updateNetworkCommands(GData);
	processCommands(GData);
	updateGameStateMessage(GData, window);
}

void updateAnimations(struct chessGData &GData){
	int controlContext = GData.controlContext;
	switch(controlContext){
		case MAIN_MENU:
			updateMenuBGAnimation(GData.menuBG.menuBoard, GData.menuBG.tileOffsets);
			break;
		default:
			break;
	}
}

void updateMenuBGAnimation(struct tile menuBoard[8][8], struct tileCoords tileOffsets[8][8]){
	int speed = 24;
	//displacement data	
	static bool moveRow = rand() % 2;
	static bool posMove = rand() % 2;
	static GLfloat curDisp = 0;
	//random index range from 2 - 6 (8 takes too long)
	static int tarIndexShift = (posMove) ? (rand() % 4 + 1) * 2 : -1 * ((rand() % 4 + 1) * 2);
	static GLfloat tarDisp = tarIndexShift * 10;
	//target line strip
	static int tarLineStrip = rand() % 8;
	//tile attributes
	static struct tileCoords defTileOffsets[8][8];
	static bool defTileLocsInitialized = false;
	if(!defTileLocsInitialized){
		initTileCoords(defTileOffsets);
		defTileLocsInitialized = true;
	}
	//timer
	static double lastTime = 0;
	double timeDisp = glfwGetTime() - lastTime;
	lastTime = glfwGetTime();
	//update displacement data
	if(posMove){
		curDisp += (timeDisp) * speed;
	}
	else if (!posMove){
		curDisp -= (timeDisp) * speed;
	}

	updateMenuOffsets(tileOffsets, defTileOffsets, curDisp, tarLineStrip, moveRow);
	//-------------------------reset parameters for new line shift if thresholds reached
	if((posMove && curDisp > tarDisp) || (!posMove && curDisp < tarDisp)){
		//apply movement changes to the menuBoard
		shiftStrip(menuBoard, tarIndexShift, tarLineStrip, moveRow);
		//update displacement data
		moveRow = rand() % 2;
		posMove = rand() % 2;
		curDisp = 0;
		tarIndexShift = (posMove) ? (rand() % 4 + 1) * 2 : -1 * ((rand() % 4 + 1) * 2);
		tarDisp = tarIndexShift * 10;
		//update line strip
		tarLineStrip = rand() % 8;
		updateMenuOffsets(tileOffsets, defTileOffsets, curDisp, tarLineStrip, moveRow);
	}
}

void updateMenuOffsets(struct tileCoords tileOffsets[8][8], struct tileCoords defTileOffsets[8][8],
	   	GLfloat disp, int tarLineStrip, bool moveRow)
{
		for(int i = 0; i < 8; ++i){
		for(int j = 0; j < 8; ++j){
			//set offsets
			//moving row
			if(moveRow){
				if(j == tarLineStrip){
					tileOffsets[i][j].x = fmod(defTileOffsets[i][j].x + disp + 80, 80);
					tileOffsets[i][j].y = defTileOffsets[i][j].y;
				}
				else if(j != tarLineStrip){
					tileOffsets[i][j].x = defTileOffsets[i][j].x;
					tileOffsets[i][j].y = defTileOffsets[i][j].y;
				}
			}
			//moving column
			else if(!moveRow){
				if(i == tarLineStrip){
					tileOffsets[i][j].x = defTileOffsets[i][j].x;
					tileOffsets[i][j].y = fmod(defTileOffsets[i][j].y - disp + 80, 80);
				}
				else if(i != tarLineStrip){
					tileOffsets[i][j].x = defTileOffsets[i][j].x;
					tileOffsets[i][j].y = defTileOffsets[i][j].y;
				}
			}
		}
	}
}


void updateCursor(struct chessGData &GData){
	GLfloat VVWidth = GData.VVPosX - GData.VVNegX;
	//the ratio of the cursor x value to the max width value
	GLfloat wRatio = (GLfloat)GData.cursorX / (GLfloat)GData.windowW;
	GData.tCursX = 	wRatio * VVWidth;

	GLfloat VVHeight = GData.VVPosY - GData.VVNegY;
	//the ratio of the cursor y value to the max height value
	GLfloat hRatio = (GLfloat)GData.cursorY / (GLfloat)GData.windowH;
	GData.tCursY = 	hRatio * VVHeight;
}

//assumes that the top left corner of the screen is always the top left corner of the board
//this lets the viewing volume add to the right and down to keep the same proportions
void updateCursorSelection(struct menuSelections &menuSel, struct tileIndex &selectedTile, 
		GLfloat tCursX, GLfloat tCursY, int controlContext, int menuState){
	switch(controlContext){	
		case GAME_TURN:
		case HOLDING_PIECE:
		case WAIT_SIDE_CHOICE:
		case WAIT_PIECE_CHOICE:
		case GAME_COMPLETION:
			//remove previous coordinates
			selectedTile.x = NOTHING;	
			selectedTile.y = NOTHING;	
			menuSel.exitToMainMenu = false;
			if(GData.inGameMenuOpen){
				if(tCursX >= 20 && tCursX <= 60 && tCursY >= 35 && tCursY <= 45){
					menuSel.exitToMainMenu = true;
				}
			}
			else if(!GData.inGameMenuOpen){
				if((tCursX >=0 && tCursX <= 80) && (tCursY >= 0 && tCursY <= 80)){
					int boardX = tCursX / 10;
					int boardY = tCursY / 10;
					if(boardX > 7)
						boardX = 7;
					if(boardX < 0)
						boardX = 0;
					if(boardY > 7)
						boardY = 7;
					if(boardY < 0)
						boardY = 0;

					selectedTile.x = boardX;
					selectedTile.y = boardY;
				}
			}
			break;
		case MAIN_MENU:
			//main
			menuSel.localPlay = false;
			menuSel.networkPlay = false;
			//network
			menuSel.client = false;
			menuSel.server = false;
			if(controlContext == MAIN_MENU){
				switch(menuState){
					case MENU_MAIN_0:
						if(tCursX >= 20 && tCursX <= 60 && tCursY >= 25 && tCursY <= 35){
							menuSel.localPlay = true;
						}
						if(tCursX >= 20 && tCursX <= 60 && tCursY >= 45 && tCursY <= 55){
							menuSel.networkPlay = true;
						}
						break;
					case MENU_NETWORK_1:
						if(tCursX >= 20 && tCursX <= 60 && tCursY >= 25 && tCursY <= 35){
							menuSel.client = true;
						}
						if(tCursX >= 20 && tCursX <= 60 && tCursY >= 45 && tCursY <= 55){
							menuSel.server = true;
						}
						break;
					default:
						break;
				}
			}
			break;
		default:
			break;
	}
}

void updateGameStateMessage(struct chessGData &GData, GLFWwindow *window){
	//static bool to prevent checking state if a new move wasnt' made
	char windowTitle[256];
	std::string networkType;
	//setup title message prefix
	switch(GData.networkD.networkRole){
		case NON_NETWORK:
			networkType = "Chess: ";
			sprintf(windowTitle, "Chess");
			break;
		case CLIENT:
			networkType = "Client";
			if(!GData.networkD.clientData.connectedToServer){
				networkType = "Client (no connection)";
			}
			networkType += GData.networkD.textMessage;
			sprintf(windowTitle, "Client%s", GData.networkD.textMessage);
			break;
		case SERVER:
			networkType = "Server";
			sprintf(windowTitle, "Server (%d connected) (White[%d], Black[%d])",
			GData.networkD.serverData.connectedHosts, whitePlayerSlotOccupied(GData.networkD.serverData)
			, blackPlayerSlotOccupied(GData.networkD.serverData));
			break;
		default:
			break;
	}
	//check, checkmate, and draw
	if(isInCheck(GData.board, GData.sMoves, GData.whiteTurn)){
		sprintf(windowTitle, "%sCheck!", networkType.c_str());
	}
	if(isCheckMate(GData.board, GData.sMoves, GData.controlContext, GData.whiteTurn)){
		if(GData.whiteTurn){
			sprintf(windowTitle, "%sCheckmate! Black wins",
			networkType.c_str());
		}
		else{
			sprintf(windowTitle, "%sCheckmate! White wins",
			networkType.c_str());
		}
	}
	else if(isDraw(GData.board, GData.sMoves, GData.controlContext, GData.whiteTurn)){
		sprintf(windowTitle, "%sDraw!", networkType.c_str());
	}
	//side selection
	if(GData.controlContext == WAIT_SIDE_CHOICE){
		if(GData.networkD.networkRole != SERVER){
			sprintf(windowTitle, "%s", networkType.c_str());
		}
	}
	//pawn promotion
	if(GData.controlContext == WAIT_PIECE_CHOICE){
		sprintf(windowTitle, "%sPawn Promotion! Press (1:Queen) (2:Rook) (3:Bishop) (4:Knight)",
			   networkType.c_str());
	}
	//apply the windowTitle
	glfwSetWindowTitle(window, windowTitle);
}

void processCommands(struct chessGData &GData){
	networkPackage curNetPack;
	command curCom;
	while(GData.netPacks.size() > 0){
		curNetPack = GData.netPacks.front();
		curCom = GData.netPacks.front().com;
		//process command with appropriate context
		switch(GData.controlContext){
			case MAIN_MENU:
				processMainMenuComs(GData, curCom);
				break;
			case GAME_TURN:
				processGameTurnComs(GData, curCom);
				break;
			case HOLDING_PIECE:
				processHoldingPieceComs(GData, curCom);
				break;
			case WAIT_SIDE_CHOICE:
				processWaitSideChoiceComs(GData, curCom);
				break;
			case WAIT_PIECE_CHOICE:
				processWaitPieceChoiceComs(GData, curCom);
				break;
			case GAME_COMPLETION:
				processGameCompletionComs(GData, curCom);
				break;
			default:
				break;
		}
		//process in-game menu commands
		processInGameMenuComs(GData, curCom);
		//check to make sure a netPack hasn't been popped during reinitialization in process
		//GameCompletionComs()
		if(GData.netPacks.size() > 0){
			GData.netPacks.pop();
		}
	}
}

void processMainMenuComs(struct chessGData &GData, struct command curCom){
	switch(curCom.type){
		case SELECT:
			if(GData.menuSel.localPlay){
				GData.controlContext = GAME_TURN;
			}
			if(GData.menuSel.networkPlay){
				GData.menuState = MENU_NETWORK_1;
			}
			if(GData.menuSel.client){
				GData.controlContext = WAIT_SIDE_CHOICE;
				initClient(GData.networkD.clientData, GData.configD.serverIP,
				GData.configD.serverPort, GData.networkD.networkRole);
			}
			if(GData.menuSel.server){
				GData.controlContext = GAME_TURN;
				initServer(GData.networkD.serverData, GData.configD.serverPort,
				GData.networkD.networkRole);
			}
			break;
		case DESELECT:
			GData.menuState = MENU_MAIN_0;
			break;
		case KEYBOARD_PRESS:
			if(curCom.choice == ESCAPE){
				GData.menuState = MENU_MAIN_0;
			}
			break;
		default:
			break;
	}
}

void processInGameMenuComs(struct chessGData &GData, struct command curCom){
	switch(GData.controlContext){
		case GAME_TURN:
		case HOLDING_PIECE:
		case WAIT_SIDE_CHOICE:
		case WAIT_PIECE_CHOICE:
		case GAME_COMPLETION:
			switch(curCom.type){
				case SELECT:
					if(GData.menuSel.exitToMainMenu){
						deInitNetworking(GData);
						initChessGData(GData);
					}
					break;
				case KEYBOARD_PRESS:
					if(curCom.choice == ESCAPE && GData.controlContext != GAME_COMPLETION){
						GData.inGameMenuOpen = !GData.inGameMenuOpen;
					}
				default:
					break;
			}
			break;
		default:
			break;
	}
}

void processGameTurnComs(struct chessGData &GData, struct command curCom){
	int x = curCom.target.x; int y = curCom.target.y;
	switch(curCom.type){
		case SELECT:
			if(isTurnPiece(GData.board[x][y].piece, GData.whiteTurn)){
				GData.heldPieceLoc.x = x; GData.heldPieceLoc.y = y;
				GData.controlContext = HOLDING_PIECE;
			}
			break;
		case DESELECT:
			break;
		default:
			break;
	}
}

void processHoldingPieceComs(struct chessGData &GData, struct command curCom){
	switch(curCom.type){
		case SELECT:
			//if you click the piece you are holding, deselect
			if(curCom.target.x == GData.heldPieceLoc.x &&
				curCom.target.y == GData.heldPieceLoc.y
			){
				GData.heldPieceLoc.x = NOTHING; GData.heldPieceLoc.y = NOTHING;
				GData.controlContext = GAME_TURN;
			}
			//else execute the move
			else{
				//if the move is complete we change turns
				if(executeMove(GData.board, GData.sMoves, GData.heldPieceLoc, 
						curCom.target, GData.controlContext, GData.whiteTurn))
				{
					changeTurn(GData.board, GData.lastBoard, GData.whiteTurn);	
				}

			}
			break;
		case DESELECT:
			GData.heldPieceLoc.x = NOTHING; GData.heldPieceLoc.y = NOTHING;
			GData.controlContext = GAME_TURN;
			break;
		default:
			break;
	}
}

void processWaitSideChoiceComs(struct chessGData &GData, struct command curCom){
	if(curCom.type == KEYBOARD_PRESS && 
	curCom.choice >= WHITE_PLAYER && curCom.choice <= BLACK_PLAYER){
		bool controlsWhite = false; bool controlsBlack = false;
		switch(curCom.choice){
			case WHITE_PLAYER:
				controlsWhite = true;
				break;
			case BLACK_PLAYER:
				controlsBlack = true;
				break;
			default:
				break;
		}
	}
}

void processWaitPieceChoiceComs(struct chessGData &GData, struct command curCom){
	//here's the check to make sure the choice is valid
	if(curCom.choice >= QUEEN && curCom.choice <= KNIGHT){
		int choice;
		switch(curCom.choice){
			case QUEEN:
				choice = QUEEN;
				break;
			case ROOK:
				choice = ROOK;
				break;
			case BISHOP:
				choice = BISHOP;
				break;
			case KNIGHT:
				choice = KNIGHT;
				break;
			default:
				break;
		}
		executePromotion(GData.board, GData.sMoves, GData.whiteTurn, choice);
		changeTurn(GData.board, GData.lastBoard, GData.whiteTurn);
		GData.controlContext = GAME_TURN;
	}
}

void processGameCompletionComs(struct chessGData &GData, struct command curCom){
}
