/*
Contributors: Liam T. Monson
License: BSD 2-Clause license which should be found in the "license.txt"
file in the root game directory.
*/
#define GLFW_INCLUDE_GLU
#include "mainChess.hpp"
using namespace std;

//--------------------------------
//global variables
GLFWwindow *window;
chessGData GData;
GLuint texObjects[NUM_TEXTURES];

void resizeWindow(GLFWwindow* window, int width, int height){
	GData.windowW = width;
	GData.windowH = height;
	//make window square
	int length = height;
	if(width < height){
		length = width;
	}
	glfwSetWindowSize(window, length, length);

	glViewport(0, 0, length, length);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	GData.VVNegX = 0.0f; GData.VVPosX = 80.0f;
	GData.VVNegY = 0.0f; GData.VVPosY = 80.0f;
	glOrtho(GData.VVNegX, GData.VVPosX, GData.VVNegY, GData.VVPosY, -10.0f, 10.0f);
}

void initializeContexts(){
	GData.windowW = GData.configD.initWinWidth;
	GData.windowH = GData.configD.initWinHeight;
	GData.cursorX = 0;
	GData.cursorY = 0;
	//-------------------------stdlib
	srand(time(NULL));	
	//--------------------------------------------------OpenGL context setup
	//-------------------------OpenGL general features
	//stops rendering of polygon backs (determined by winding)
	glEnable(GL_CULL_FACE);	
	//enables testing each fragment's depth vs the previous depth in the buffer
	glEnable(GL_DEPTH_TEST);
	//prevents alpha channels lower than 0.01 from being written to the depth/color buffers
	glAlphaFunc(GL_GREATER, 0.01f);
	glEnable(GL_ALPHA_TEST);
	//-------------------------enable/initialize texture mapping
	glEnable(GL_TEXTURE_2D);
	glGenTextures(NUM_TEXTURES, texObjects);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glColor3f(0.0f, 1.0f, 0.0f);
	//allows for correct alpha blending
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	for(int i = 0; i < NUM_TEXTURES; ++i){
		GLint x, y, comp;
		GLubyte *data;
		FILE *file = fopen(textureFiles[i], "rb");
		data = stbi_load_from_file(file, &x, &y, &comp, 0);
		fclose(file);

		glBindTexture(GL_TEXTURE_2D, texObjects[i]);
		//stores pixels on OpenGL implementation with given alignment
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		// non-mipmap
		/*
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		*/
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		if( comp == 3 ) {
			flipImage(data, x, y, comp);
			glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, x, y, 0, GL_RGB, GL_UNSIGNED_BYTE, data );
		} else if( comp == 4 ) {
			flipImage(data, x, y, comp);
			glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, x, y, 0, GL_RGBA, GL_UNSIGNED_BYTE, data );
		}

		stbi_image_free(data);
	}
	//--------------------------------------------------chess board initialization
	initChessGData(GData);
}

void errorCallback(int error, const char* description){
//	fprintf(stderr, "Error: %s\n", description);
}

int main(int argc, char* argv[]){
	//grab information from config file
	parseConfig(GData);
	//-------------------------GLFW code
	glfwSetErrorCallback(errorCallback);

	if(!glfwInit()){
		exit(EXIT_FAILURE);
	}

	window = glfwCreateWindow(GData.configD.initWinWidth, GData.configD.initWinHeight, 
			"Chess implemented by Liam Monson", NULL, NULL);
	if(!window){
		glfwTerminate();
		exit(EXIT_FAILURE);
	}

	glfwSetFramebufferSizeCallback(window, resizeWindow);
	glfwSetKeyCallback(window, handleKeys);
	glfwSetMouseButtonCallback(window, handleMouseButton);

	//glfw context must be current before applying OpenGL context modifications
	glfwMakeContextCurrent(window);
	initializeContexts();

	glfwSwapInterval(0);
	
	resizeWindow(window, GData.configD.initWinWidth, GData.configD.initWinHeight);
	
	while(!glfwWindowShouldClose(window)){
		glfwPollEvents();
		glfwGetCursorPos(window, &GData.cursorX, &GData.cursorY);
		updateGame(GData, window);
		render(GData); 
	}

	glfwTerminate();
	cleanUp(GData);
	exit(EXIT_SUCCESS);
}

void render(struct chessGData &GData){
	//setup FPS regulation variables
	static double renderTime = 0.0;
	static double lastFrame;
	lastFrame = glfwGetTime() - renderTime;
	//if less than a 250 of a second has passed since the last frame
	//cpu throttling
	int sleepMilli = (int) (((1.0 / MAX_FPS) - lastFrame) * 1000 + 0.5);
	if(lastFrame < 1.0 / MAX_FPS){
		std::this_thread::sleep_for(std::chrono::milliseconds(sleepMilli));
	}
	//if longer than a 250 of a second has passed since the last frame
	if(lastFrame > 1.0 / MAX_FPS){
		renderTime = glfwGetTime();
		//--------------------------------------------------clear and setup context
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		//render appropriate data
		switch(GData.controlContext){
			case MAIN_MENU:
				renderMenu(GData.menuBG, GData.menuSel, GData.menuState);
				break;
			case GAME_TURN:
			case HOLDING_PIECE:
			case WAIT_SIDE_CHOICE:
			case WAIT_PIECE_CHOICE:
			case GAME_COMPLETION:
				renderGameBoard(GData);
				renderInGameMenu(GData.menuSel, GData.inGameMenuOpen);
				break;
			default:
				break;
		}
		/*
		//line crosshair, keeping for easy start on custom cursor later
		glColor3f(0.0f, 0.0f, 1.0f);
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_LINES);
			//horizontal
			glVertex3f(0.0f, GData.VVPosY - GData.tCursY, 9.0f);
			glVertex3f(GData.VVPosX, GData.VVPosY - GData.tCursY, 9.0f);
			//vertical
			glVertex3f(GData.tCursX, GData.VVNegY, 9.0f);
			glVertex3f(GData.tCursX, GData.VVPosY, 9.0f);
		glEnd();
		*/
		glfwSwapBuffers(window);
	}
}

void renderMenu(struct menuBackground &menuBG, struct menuSelections menuSel, int menuState){
	tile (&menuBoard)[8][8] = menuBG.menuBoard;
	GLfloat xOffset = 0;
	GLfloat yOffset = 0;
	//-------------------------Render background
	//render tile animation
	glEnable(GL_TEXTURE_2D);
	glColor3f(1.0f, 0.0f, 0.0f);
	for(int i = 0; i < 8; ++i){
		for(int j = 0; j < 8; ++j){
			xOffset = menuBG.tileOffsets[i][j].x; yOffset = menuBG.tileOffsets[i][j].y;
			glPushMatrix();
			glTranslatef(xOffset, yOffset, 0);
			renderTile(menuBoard[i][j], true, true);
			glPopMatrix();
			//render fill tiles
			if(xOffset < 5 || xOffset > 75){ 
				//row fill tiles
				glPushMatrix();
				glTranslatef(xOffset + 80.0f, yOffset, 0);
				renderTile(menuBoard[i][j], true, true);
				glPopMatrix();	
				glPushMatrix();
				glTranslatef(xOffset - 80.0f, yOffset, 0);
				renderTile(menuBoard[i][j], true, true);
				glPopMatrix();	
			}
			if(yOffset < 5 || yOffset > 75){
				//column fill tiles
				glPushMatrix();
				glTranslatef(xOffset, yOffset + 80.0f, 0);
				renderTile(menuBoard[i][j], true, true);
				glPopMatrix();	
				glPushMatrix();
				glTranslatef(xOffset, yOffset - 80.0f, 0);
				renderTile(menuBoard[i][j], true, true);
				glPopMatrix();	
			}
		}
	}
	//-------------------------render buttons
	switch(menuState){
		case MENU_MAIN_0:
			//local play
			glPushMatrix();
			glTranslatef(40, 50, 0);
			renderMenuButton(LOCAL_PLAY, menuSel.localPlay);
			glPopMatrix();
			//network play
			glPushMatrix();
			glTranslatef(40, 30, 0);
			renderMenuButton(NETWORK_PLAY, menuSel.networkPlay);
			glPopMatrix();
			//render chess title
			glPushMatrix();
			glTranslatef(40, 70, 0);
			renderChessTitle();
			glPopMatrix();
			break;
		case MENU_NETWORK_1:
			//create client
			glPushMatrix();
			glTranslatef(40, 50, 0);
			renderMenuButton(NETWORK_CLIENT, menuSel.client);
			glPopMatrix();
			//create server
			glPushMatrix();
			glTranslatef(40, 30, 0);
			renderMenuButton(NETWORK_SERVER, menuSel.server);
			glPopMatrix();
			break;
		default:
			break;
	}
	
	glDisable(GL_TEXTURE_2D);
}

void renderInGameMenu(struct menuSelections &menuSel, bool inGameMenuOpen){
	if(inGameMenuOpen){
		glPushMatrix();
		glTranslatef(40, 40, 0);
		renderMenuButton(GO_TO_MAIN_MENU, menuSel.exitToMainMenu);
		glPopMatrix();
	}
}

void renderGameBoard(struct chessGData &GData){
	//tile attributes
	static struct tileCoords tileLocs[8][8];
	static bool tileLocsInitialized = false;
	if(!tileLocsInitialized){
		initTileCoords(tileLocs);
		tileLocsInitialized = true;
	}
	int tarX = GData.selectedTile.x;
	int tarY = GData.selectedTile.y;
	tileIndex heldLoc = GData.heldPieceLoc;
	int controlContext = GData.controlContext;
	glColor3f(1.0f, 1.0f, 1.0f);
	glEnable(GL_TEXTURE_2D);
	//render tiles
	for(int i = 0; i < 8; ++i){
		for(int j = 0; j < 8; ++j){
			glPushMatrix();
			glTranslatef(tileLocs[i][j].x, tileLocs[i][j].y, 0);
			renderTile(GData.board[i][j], !isHeldPiece(heldLoc, i, j) && 
				controlContext == HOLDING_PIECE, GData.whiteTurn);
			glPopMatrix();
		}
	}
	//render highlight
	if(!(GData.selectedTile.x == NOTHING || GData.selectedTile.y == NOTHING)){
		glPushMatrix();
		glTranslatef(tileLocs[tarX][tarY].x, tileLocs[tarX][tarY].y, 0);
		renderHighlight(GData.board[tarX][tarY], GData.whiteTurn);
		glPopMatrix();
	}
	
	glDisable(GL_TEXTURE_2D);
}

void renderChessTitle(){
	GLfloat width = 40;
	GLfloat height = 10;
	GLfloat halfWidth = 0.5 * width;
	GLfloat halfHeight = 0.5 * height;
	GLfloat z = 9;
	//black background
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	glBegin(GL_TRIANGLE_FAN);
	glColor4f(0.0f, 0.0f, 0.0f, 0.9f);
		glVertex3f(halfWidth, -halfHeight, z - 1);
		glVertex3f(halfWidth, halfHeight, z - 1);
		glVertex3f(-halfWidth, halfHeight, z - 1);
		glVertex3f(-halfWidth, -halfHeight, z - 1);
	glEnd();
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texObjects[CHESS_TITLE]);
	//text
	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_TRIANGLE_FAN);
		glTexCoord2f(1.0f, 0.0f);
		glVertex3f(halfWidth, -halfHeight, z);
		glTexCoord2f(1.0f, 1.0f);
		glVertex3f(halfWidth, halfHeight, z);
		glTexCoord2f(0.0f, 1.0f);
		glVertex3f(-halfWidth, halfHeight, z);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(-halfWidth, -halfHeight, z);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
};

void renderMenuButton(int button, bool isSelected){
	GLfloat width = 40;
	GLfloat height = 10;
	GLfloat halfWidth = 0.5 * width;
	GLfloat halfHeight = 0.5 * height;
	GLfloat halfWidthPlus = halfWidth + 1;
	GLfloat halfHeightPlus = halfHeight + 1;
	GLfloat z = 8;
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	//black background or selection
	if(isSelected){
		glColor4f(0.38f, 0.0f, 0.71f, 0.9f);
	}
	else{
		glColor4f(0.0f, 0.0f, 0.0f, 0.9f);
	}
	glBegin(GL_TRIANGLE_FAN);
		glVertex3f(halfWidth, -halfHeight, z - 1.5);
		glVertex3f(halfWidth, halfHeight, z - 1.5);
		glVertex3f(-halfWidth, halfHeight, z - 1.5);
		glVertex3f(-halfWidth, -halfHeight, z - 1.5);
	glEnd();
	//text
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texObjects[button]);
	glBegin(GL_TRIANGLE_FAN);
		glTexCoord2f(1.0f, 0.0f);
		glVertex3f(halfWidth, -halfHeight, z);
		glTexCoord2f(1.0f, 1.0f);
		glVertex3f(halfWidth, halfHeight, z);
		glTexCoord2f(0.0f, 1.0f);
		glVertex3f(-halfWidth, halfHeight, z);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(-halfWidth, -halfHeight, z);
	glEnd();
	glEnable(GL_DEPTH_TEST);
}

void renderTile(struct tile tarTile, bool skipGlow, bool whiteTurn){
	int color = tarTile.color;
	int piece = tarTile.piece;
	int moveType = tarTile.moveType;
	bool lastMove = tarTile.lastMove;
	//----render base tile
	glBindTexture(GL_TEXTURE_2D, texObjects[color]);
	renderTSquare(10.0f, (GLfloat) 0.0f);
	//----render glow
	//turn color
	if(isTurnPiece(piece, whiteTurn) && !skipGlow){
		if(whiteTurn){
			glBindTexture(GL_TEXTURE_2D, texObjects[WGLOW]);
		}
		else{
			glBindTexture(GL_TEXTURE_2D, texObjects[BGLOW]);
		}
		renderTSquare(10.0f, (GLfloat) 1.0f);
	}
	//----render move type 
	if(lastMove == true && !skipGlow){
		glBindTexture(GL_TEXTURE_2D, texObjects[LAST_MOVE]);
		renderTSquare(10.0f, (GLfloat) 1.15f);
	}
	if(isValidMove(moveType)){
		glBindTexture(GL_TEXTURE_2D, texObjects[VALID_MOVE]);
		renderTSquare(10.0f, (GLfloat) 1.2f);
	}
	if(isInvalidMove(moveType)){
		glBindTexture(GL_TEXTURE_2D, texObjects[INVALID_MOVE]);
		renderTSquare(10.0f, (GLfloat) 1.3f);
	}
	//render piece
	if(tarTile.piece != NOTHING){
		glBindTexture(GL_TEXTURE_2D, texObjects[tarTile.piece]);
		renderTSquare(10.0f, (GLfloat) 2.0f);
	}
}

void renderHighlight(struct tile tarTile, bool whiteTurn){
	int piece = tarTile.piece;
	int moveType = tarTile.moveType;
	if(isTurnPiece(piece, whiteTurn)){
		if(whiteTurn){
			glBindTexture(GL_TEXTURE_2D, texObjects[S_W_HIGHLIGHT]);
		}
		else{
			glBindTexture(GL_TEXTURE_2D, texObjects[S_B_HIGHLIGHT]);
		}
	}
	else{
		glBindTexture(GL_TEXTURE_2D, texObjects[MOVE_HIGHLIGHT]);
	}
	renderTSquare(10.0f, (GLfloat) 1.5f);
}

void renderTSquare(GLfloat size, GLfloat z){
	GLfloat halfSize = 0.5 * size;
	glBegin(GL_TRIANGLE_FAN);
		glTexCoord2f(1.0f, 0.0f);
		glVertex3f(halfSize, -halfSize, z);
		glTexCoord2f(1.0f, 1.0f);
		glVertex3f(halfSize, halfSize, z);
		glTexCoord2f(0.0f, 1.0f);
		glVertex3f(-halfSize, halfSize, z);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(-halfSize, -halfSize, z);
	glEnd();
}
