/*
Contributors: Liam T. Monson
License: BSD 2-Clause license which should be found in the "license.txt"
file in the root game directory.
*/
#include "networking.hpp"
void initNetworking(chessGData &GData, int &networkRole){
	static bool initialized = false;
	if(!initialized){
		enet_initialize();
	}
	networkRole = NON_NETWORK;
}

void deInitNetworking(chessGData &GData){
	networkData *networkD = &GData.networkD;
	//disconnect from any peers
	//if running as client
	if(networkD->networkRole == CLIENT){
		//if connected
		if(networkD->clientData.connectedToServer){
			enet_peer_disconnect_later(networkD->clientData.peer, 0);
			enet_host_service(networkD->clientData.client, &GData.networkD.clientData.event, 0);
		}
	}
	//if running as server
	if(networkD->networkRole == SERVER){
		for(int i = 0; i < MAX_SERVER_PEER_COUNT; ++i){
			enet_peer_disconnect_later(&networkD->serverData.server->peers[i], 0);
		}
		enet_host_service(networkD->serverData.server, &networkD->serverData.event, 0);
	}
	//shutdown enet
	enet_deinitialize();
}

void updateNetworkCommands(chessGData &GData){
	networkData *networkD = &GData.networkD;
	switch(GData.networkD.networkRole){
		case CLIENT:
			updateClientCommands(GData, networkD->clientData);
			break;
		case SERVER:
			updateServerCommands(GData, networkD->serverData);
			break;
		default:
			break;
	}
}

void initNetPack(chessGData &GData, networkPackage &newNetPack, command newCommand, packageType packType){
	newNetPack.packType = packType;
	newNetPack.state.context = GData.controlContext;
	newNetPack.com = newCommand;
	for(int i = 0; i < NETWORK_TEXT_MESSAGE_CHARS; ++i){
		newNetPack.textMessage[i] = '\0';
	}
}

//--------------------------------------------------client
void initClient(dataForClient &clientData, std::string serverIP, enet_uint16 serverPort, int &networkRole){
	networkRole = CLIENT;
	initClientAddress(clientData.address, serverIP, serverPort);
	initClientHost(clientData);
	connectClientPeer(clientData);
	clientData.connectedToServer = false;
}

void initClientAddress(ENetAddress &address, std::string serverIP, enet_uint16 serverPort){
    enet_address_set_host(&address, serverIP.c_str());
    address.port = serverPort;
}

void initClientHost(dataForClient &clientData){
	clientData.client = enet_host_create(NULL, 1, 2, 57600/8, 14400/8);
}

void connectClientPeer(dataForClient &clientData){
	clientData.peer = enet_host_connect(clientData.client, &clientData.address, 2, 0);
}

void updateClientCommands(chessGData &GData, dataForClient &clientData){
	sendClientPackages(GData, clientData);
	//recieve server commands
	while(enet_host_service(clientData.client, &clientData.event, 0) > 0){
		switch(clientData.event.type){
			case ENET_EVENT_TYPE_CONNECT:
				clientData.connectedToServer = true;
				break;
			case ENET_EVENT_TYPE_RECEIVE:
				processClientPackages(GData, clientData);
				break;
			case ENET_EVENT_TYPE_DISCONNECT:
				clientData.connectedToServer = false;
				clientData.event.peer->data = NULL;
				break;
		}
	}
}

void updateNetworkTextMessage(networkData &networkD, networkPackage package){
	for(int i = 0; i < NETWORK_TEXT_MESSAGE_CHARS; ++i){
		networkD.textMessage[i] = package.textMessage[i];
	}
}

void processClientPackages(chessGData &GData, dataForClient &clientData){
	networkPackage receivedPackage;
	packageType receivedType;
	//grab packType
	memcpy(&receivedType, clientData.event.packet->data, sizeof(packageType));
	switch(receivedType){
		case ANNOUNCEMENT:
		memcpy(&receivedPackage, clientData.event.packet->data, sizeof(networkPackage));
		updateNetworkTextMessage(GData.networkD, receivedPackage);
		GData.controlContext = receivedPackage.state.context;
		enet_packet_destroy(clientData.event.packet);
			break;
		case COMMAND:
		memcpy(&receivedPackage, clientData.event.packet->data, sizeof(networkPackage));
		updateNetworkTextMessage(GData.networkD, receivedPackage);
		//update context with server context
		GData.controlContext = receivedPackage.state.context;
		GData.netPacks.push(receivedPackage);
		enet_packet_destroy(clientData.event.packet);
			break;
		case FULL_GAME_UPDATE:
		memcpy(&receivedPackage, clientData.event.packet->data, sizeof(networkPackage));
		processFullUpdatePackage(GData, clientData);
			break;
		default:
			break;
	}
}

void processFullUpdatePackage(chessGData &GData, dataForClient &clientData){
	fullUpdatePackage receivedPackage;
	memcpy(&receivedPackage, clientData.event.packet->data, sizeof(fullUpdatePackage));
	GData.sMoves = receivedPackage.sMoves;
	GData.heldPieceLoc = receivedPackage.heldPieceLoc;
	for(int i = 0; i < 8; ++i){
		for(int j = 0; j < 8; ++j){
			GData.board[i][j] = receivedPackage.board[(i * 8) + j];
			GData.lastBoard[i][j] = receivedPackage.board[(i * 8) + j];
		}
	}
	GData.whiteTurn = receivedPackage.whiteTurn;
}

void sendClientPackages(chessGData &GData, dataForClient &clientData){
	ENetPacket *packet;
	networkPackage toServer;
	std::queue<networkPackage> localNetPacks;
	while(GData.netPacks.size() > 0 && clientData.connectedToServer){
		//if the command is meant for sending pop it off and send it
		if(GData.netPacks.front().packType != LOCAL_ONLY){
			if(GData.controlContext != WAIT_SIDE_CHOICE){
				toServer.packType = COMMAND;
				toServer = GData.netPacks.front();
				packet = 
					enet_packet_create(&toServer, sizeof(networkPackage), ENET_PACKET_FLAG_RELIABLE);
				enet_peer_send(clientData.peer, 0, packet);
			}
			//else we're making an announcement
			else{
				toServer.packType = ANNOUNCEMENT;
				toServer.com = GData.netPacks.front().com;
				packet = 
					enet_packet_create(&toServer, sizeof(networkPackage), ENET_PACKET_FLAG_RELIABLE);
				enet_peer_send(clientData.peer, 0, packet);
			}
		}
		//if the command should stay local dump it to temporary storage to add back to the netPack queue
		//after all the netPacks have been handled
		if(GData.netPacks.front().packType == LOCAL_ONLY){
			localNetPacks.push(GData.netPacks.front());
		}
		GData.netPacks.pop();
	}
	//now that all netPacks have been either sent or put back on the localNetPacks queue
	//add the local netPacks back for local processing
	while(localNetPacks.size() >0){
		GData.netPacks.push(localNetPacks.front());
		localNetPacks.pop();
	}
}

//--------------------------------------------------server
void initServer(dataForServer &serverData, enet_uint16 serverPort, int &networkRole){
	networkRole = SERVER;
	initServerAddress(serverData.address, serverPort);
	initServerHost(serverData);
	for(int i = 0; i < MAX_SERVER_PEER_COUNT; ++i){
		initPlayerData(serverData.playerD[i]);
	}
	serverData.connectedHosts = 0;
}

void initServerAddress(ENetAddress &address, enet_uint16 serverPort){
	address.host = ENET_HOST_ANY;
    address.port = serverPort;
}

void initServerHost(dataForServer &serverD){
    serverD.server = enet_host_create(&serverD.address, MAX_SERVER_PEER_COUNT, 2, 0, 0);
}

void updateServerCommands(chessGData &GData, dataForServer &serverData){
	handlePlayerSlots(GData, serverData);
	//receive commands from players and broadcast commands
	while(enet_host_service(serverData.server, &serverData.event, 0) > 0){
		switch(serverData.event.type){
			case ENET_EVENT_TYPE_CONNECT:
				setConnectPeerData(GData, serverData);
				break;
			case ENET_EVENT_TYPE_RECEIVE:
				processServerPackages(GData, serverData);
				break;
			case ENET_EVENT_TYPE_DISCONNECT:
				setDisconnectPeerData(serverData);
				break;
		}
	}
}

void handlePlayerSlots(chessGData &GData, dataForServer &serverData){
	if(!bothPlayerSlotsOccupied(serverData)){
		if(GData.controlContext != WAIT_SIDE_CHOICE){
			swapControlContextWaitForPlayers(GData.controlContext);
		}
		//sleep to prevent spamming
		std::this_thread::sleep_for(std::chrono::milliseconds(100));
		broadcastPlayerSlotChoices(GData, serverData);
	}
	else if(GData.controlContext == WAIT_SIDE_CHOICE){
		swapControlContextWaitForPlayers(GData.controlContext);
	}
}

void processServerPackages(chessGData &GData, dataForServer &serverData){
	ENetPacket *broadcastPacket;
	std::string newMessage;
	networkPackage receivedPackage;
	memcpy(&receivedPackage, serverData.event.packet->data, sizeof(networkPackage));
	switch(receivedPackage.packType){
		case ANNOUNCEMENT:
			switch(GData.controlContext){
				case WAIT_SIDE_CHOICE:
					switch(receivedPackage.com.choice){
						case WHITE_PLAYER:
							if(!whitePlayerSlotOccupied(serverData)){
								(*(playerData*) serverData.event.peer->data).playType = WHITE_PLAYER;
							}
							break;
						case BLACK_PLAYER:
							if(!blackPlayerSlotOccupied(serverData)){
								(*(playerData*) serverData.event.peer->data).playType = BLACK_PLAYER;
							}
							break;
						default:
							break;
					}
					if(bothPlayerSlotsOccupied(serverData)){
						swapControlContextWaitForPlayers(GData.controlContext);
						broadcastServerAnnouncement(GData, serverData, GData.controlContext);
					}
				default:
					break;
			}
			break;
		case COMMAND:
			//if the appropriate player is giving commands
			if(appropriateCommandAuthority(serverData, GData.controlContext))
			{
				//add command to server commands for processing
				GData.netPacks.push(receivedPackage);
				//make sure the server's state is being enforced
				setServerStateEnforcement(GData, receivedPackage);
				//broadcast
				for(int i = 0; i <= MAX_SERVER_PEER_COUNT; ++i){
					setNetworkTextMessage(receivedPackage, serverData.playerD[i].playType);
					broadcastPacket = 
					enet_packet_create(&receivedPackage, sizeof(networkPackage),
					ENET_PACKET_FLAG_RELIABLE);
					enet_peer_send(&serverData.server->peers[i], 0, broadcastPacket);
				}
			}
			break;
		default:
			break;
	}
	enet_packet_destroy(serverData.event.packet);
}

void initPlayerData(playerData &playerD){
	playerD.peerInfo = NULL;
	playerD.playType = OBSERVER;
	playerD.connected = false;	
}

void setPlayerData(playerData &playerD, ENetPeer *peerInfo, playerType type, bool connected){
	playerD.peerInfo = peerInfo;
	playerD.playType = type;
	playerD.connected = connected;	
}

void setConnectPeerData(chessGData &GData, dataForServer &serverData){
	for(int i = 0; i < MAX_SERVER_PEER_COUNT; ++i){
		if(!serverData.playerD[i].connected){
			setPlayerData(serverData.playerD[i], serverData.event.peer, OBSERVER, true);
			serverData.event.peer->data = &serverData.playerD[i];
			++serverData.connectedHosts;
			break;
		}
	}
	//TODO make this actually work
	sendFullGameUpdate(GData, serverData);
}

void setDisconnectPeerData(dataForServer &serverData){
	initPlayerData(*(playerData*) serverData.event.peer->data);
	serverData.event.peer->data = NULL;
	--serverData.connectedHosts;
}

void setNetworkTextMessage(networkPackage &package, playerType type){
	std::string serverMessage = " (";
	switch(type){
		case OBSERVER:
			serverMessage += "Observer";
			break;
		case WHITE_PLAYER:
			serverMessage += "White";
			break;
		case BLACK_PLAYER:
			serverMessage += "Black";
			break;
		default:
			break;
	}
	serverMessage += "): ";
	memcpy(&package.textMessage, serverMessage.c_str(), serverMessage.size());
	//set all non-string values to NULL
	for(int i = serverMessage.size(); i < NETWORK_TEXT_MESSAGE_CHARS; ++i){
		package.textMessage[i] = '\0';
	}
}

void setServerStateEnforcement(chessGData GData, networkPackage &package){
	package.state.context = GData.controlContext;
}

bool bothPlayerSlotsOccupied(dataForServer &serverData){
	bool returnVal = false;
	if(whitePlayerSlotOccupied(serverData) && 
		blackPlayerSlotOccupied(serverData))
	{
		returnVal = true;
	}	
	return returnVal;
}

bool whitePlayerSlotOccupied(dataForServer &serverData){
	bool returnVal = false;
	for(int i = 0; i < MAX_SERVER_PEER_COUNT; ++i){
		if(serverData.playerD[i].playType == WHITE_PLAYER){
			returnVal = true;
		}
	}
	return returnVal;
}

bool blackPlayerSlotOccupied(dataForServer &serverData){
	bool returnVal = false;
	for(int i = 0; i < MAX_SERVER_PEER_COUNT; ++i){
		if(serverData.playerD[i].playType == BLACK_PLAYER){
			returnVal = true;
		}
	}
	return returnVal;
}

bool appropriateCommandAuthority(dataForServer &serverData, int controlContext){
	bool returnVal = false;
	int playType = (*(playerData*) serverData.event.peer->data).playType;
	if(((playType == WHITE_PLAYER && GData.whiteTurn) || (playType == BLACK_PLAYER && !GData.whiteTurn))
	|| (controlContext == GAME_COMPLETION && (playType == WHITE_PLAYER || playType == BLACK_PLAYER)))
	{
		returnVal = true;
	}
	return returnVal;
}

void broadcastPlayerSlotChoices(chessGData GData, dataForServer &serverData){
	ENetPacket *broadcastPacket;
	networkPackage announcePack;
	std::string serverAnnouncement = " (Choose a side! Press ";
	if(!bothPlayerSlotsOccupied(serverData)){
		//white player slot
		if(whitePlayerSlotOccupied(serverData)){
			serverAnnouncement += "(White[X])";
		}	
		else{
			serverAnnouncement += "(1: White[ ])";
		}
		//black player slot
		if(blackPlayerSlotOccupied(serverData)){
			serverAnnouncement += "(Black[X])";
		}	
		else{
			serverAnnouncement += "(2: Black[ ])";
		}
		serverAnnouncement += ")";
		//----------set data
		announcePack.packType = ANNOUNCEMENT;
		announcePack.state.context = GData.controlContext;
		//set all non-string values to NULL
		for(int i = serverAnnouncement.size(); i < NETWORK_TEXT_MESSAGE_CHARS; ++i){
			announcePack.textMessage[i] = '\0';
		}
		//broadcast announcement packet
		for(int i = 0; i < MAX_SERVER_PEER_COUNT; ++i){
			memcpy(&announcePack.textMessage, serverAnnouncement.c_str(), serverAnnouncement.size());
			//enforce server state
			setServerStateEnforcement(GData, announcePack);
			broadcastPacket = 
			enet_packet_create(&announcePack, sizeof(networkPackage),
			ENET_PACKET_FLAG_RELIABLE);
			enet_peer_send(&serverData.server->peers[i], 0, broadcastPacket);
		}
	}
}

void broadcastServerAnnouncement(chessGData GData, dataForServer &serverData, int controlContext){
	ENetPacket *broadcastPacket;
	networkPackage announcePack;
	//----------set data
	announcePack.packType = ANNOUNCEMENT;
	for(int i = 0; i < MAX_SERVER_PEER_COUNT; ++i){
		setNetworkTextMessage(announcePack, serverData.playerD[i].playType);
		//enforce server state
		setServerStateEnforcement(GData, announcePack);
		broadcastPacket = 
		enet_packet_create(&announcePack, sizeof(networkPackage),
		ENET_PACKET_FLAG_RELIABLE);
		enet_peer_send(&serverData.server->peers[i], 0, broadcastPacket);
	}
}

void sendFullGameUpdate(chessGData &Gdata, dataForServer &serverData){
	ENetPacket *packet;
	fullUpdatePackage updatePack;
	/*
	//----------set data
	updatePack.packType = COMMAND;
	setNetworkTextMessage(updatePack, (*(playerData*) serverData.event.peer->data).playType);
	for(int i = 0; i < GData.netPackLog.size(); ++i){
		updatePack = GData.netPackLog[i];
		packet = 
		enet_packet_create(&updatePack, sizeof(networkPackage),
		ENET_PACKET_FLAG_RELIABLE);
		enet_peer_send(serverData.event.peer, 0, packet);
	}
	*/
	//initialize updatePack
	updatePack.packType = FULL_GAME_UPDATE;
	updatePack.sMoves = GData.sMoves;
	updatePack.heldPieceLoc = GData.heldPieceLoc;
	for(int i = 0; i < 8; ++i){
		for(int j = 0; j < 8; ++j){
			updatePack.board[(i * 8) + j] = GData.board[i][j];
		}
	}
	updatePack.whiteTurn = GData.whiteTurn;
	packet = 
	enet_packet_create(&updatePack, sizeof(fullUpdatePackage),
	ENET_PACKET_FLAG_RELIABLE);
	enet_peer_send(serverData.event.peer, 0, packet);
}

void swapControlContextWaitForPlayers(int &controlContext){
	static int pausedContext = controlContext;		
	if(controlContext == WAIT_SIDE_CHOICE){
		controlContext = pausedContext;
	}	
	else{
		pausedContext = controlContext;
		controlContext = WAIT_SIDE_CHOICE;
	}
}
